﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Windows.Devices.Enumeration;
using Windows.Devices.Gpio;
using Windows.Devices.Spi;

namespace GoBusExperimental
{
    public class GoPort : GoPortBase, IDisposable
    {
        private bool _disposed;
        private bool _initialized;
        private readonly int _gpioPinNumber;
        private readonly string _spiName;
        private readonly int _spiChipSelectPinNumber;
        private readonly int _altSpiChipSelectPinNumber;

        public GoPort(int id, int gpioPinNumber, string spiName, int spiChipSelectPinNumber, int altSpiChipSelectPinNumber = -1)
        {
            ID = id;
            _gpioPinNumber = gpioPinNumber;
            _spiName = spiName;
            _spiChipSelectPinNumber = spiChipSelectPinNumber;
            _altSpiChipSelectPinNumber = altSpiChipSelectPinNumber;
        }

        public static async Task<GoPort> GetGoPortAsync(int id, int gpioPinNumber, string spiName, int spiChipSelectPinNumber, int altSpiChipSelectPinNumber = -1)
        {
            var port = new GoPort(id, gpioPinNumber, spiName, spiChipSelectPinNumber, altSpiChipSelectPinNumber);
            await port.InitializeAsync().ConfigureAwait(false);
            return port;
        }

        protected async Task InitializeAsync()
        {
            // TODO: This is not thread safe.
            if (_initialized)
                return;
            _initialized = true;

            // Initialize SPI
            var spiSelector = SpiDevice.GetDeviceSelector(_spiName);
            var deviceInfo = await DeviceInformation.FindAllAsync(spiSelector).AsTask().ConfigureAwait(false);
            _spiEnumeration = await SpiDevice.FromIdAsync(deviceInfo[0].Id, new SpiConnectionSettings(_spiChipSelectPinNumber) { ClockFrequency = (int)GoBus.SPIEnumerationClockRate * 1000, SharingMode = SpiSharingMode.Shared, Mode = SpiMode.Mode1 }).AsTask().ConfigureAwait(false);
            _spi = await SpiDevice.FromIdAsync(deviceInfo[0].Id, new SpiConnectionSettings(_spiChipSelectPinNumber) { ClockFrequency = (int)GoBus.SPIEnumerationClockRate * 1000, SharingMode = SpiSharingMode.Shared, Mode = SpiMode.Mode1 }).AsTask().ConfigureAwait(false);

            if (_altSpiChipSelectPinNumber >= 0)
            {
                var gpioController = GpioController.GetDefault();
                _spiCSPin = gpioController.OpenPin(_altSpiChipSelectPinNumber);
                _spiCSPin.SetDriveMode(GpioPinDriveMode.Output);
                _spiCSPin.Write(GpioPinValue.High);
            }
        }

        public override void Initialize()
        {
            base.Initialize();

            // TODO: Fix deadlock
            InitializeAsync().Wait();
        }

        protected override void OnBinding(GoModule module)
        {
            base.OnBinding(module);

            // Initialize the GPIO pin
            InitializeGpioPin();

            // Set up SPI device
            _spi.ConnectionSettings.ClockFrequency = (int)module.SPIClockRate * 1000;

            // Does this module need UART?
            var uartModule = module as IUartGoModule;
            if (uartModule != null)
            {
                // TODO
                throw new NotImplementedException();
            }
        }

        protected override void OnUnbinding()
        {
            base.OnUnbinding();

            // Release the GPIO pin
            ReleaseGpioPin();
        }

        #region GPIO

        protected GpioPin _gpioPin;

        private GoPortGpioPinMode _gpioPinMode = GoPortGpioPinMode.None;
        public override GoPortGpioPinMode GpioPinMode
        {
            get { return _gpioPinMode; }
            set
            {
                if (_gpioPinMode == value)
                    return;
                _gpioPinMode = value;
                InitializeGpioPin();
            }
        }

        private bool _gpioPinGlitchFilter;
        public override bool GpioPinGlitchFilter
        {
            get { return _gpioPinGlitchFilter; }
            set
            {
                if (_gpioPinGlitchFilter == value)
                    return;
                _gpioPinGlitchFilter = value;
                InitializeGpioPin();
            }
        }

        private PortResistorMode _gpioPinResistorMode = PortResistorMode.PullUp;
        public override PortResistorMode GpioPinResistorMode
        {
            get { return _gpioPinResistorMode; }
            set
            {
                if (_gpioPinResistorMode == value)
                    return;
                _gpioPinResistorMode = value;
                InitializeGpioPin();
            }
        }

        private PortInterruptMode _gpioPinInterruptMode = PortInterruptMode.InterruptEdgeLow;
        public override PortInterruptMode GpioPinInterruptMode
        {
            get { return _gpioPinInterruptMode; }
            set
            {
                if (_gpioPinInterruptMode == value)
                    return;
                _gpioPinInterruptMode = value;
                InitializeGpioPin();
            }
        }

        public override bool GpioRead()
        {
            return (_gpioPin.Read() == GpioPinValue.High);
        }

        public override void GpioWrite(bool value)
        {
            _gpioPin.Write(value ? GpioPinValue.High : GpioPinValue.Low);
        }

        protected virtual void ReleaseGpioPin()
        {
            if (_gpioPin != null)
            {
                _gpioPin.ValueChanged -= GpioPin_ValueChanged;
                _gpioPin.Dispose();
                _gpioPin = null;
            }
        }

        protected virtual void InitializeGpioPin()
        {
            ReleaseGpioPin();

            if (GpioPinMode == GoPortGpioPinMode.None)
                return;

            var gpioController = GpioController.GetDefault();
            var pin = gpioController.OpenPin(_gpioPinNumber);

            switch (GpioPinMode)
            {
                case GoPortGpioPinMode.Output:
                    pin.SetDriveMode(GpioPinDriveMode.Output);
                    break;

                case GoPortGpioPinMode.Input:
                case GoPortGpioPinMode.IRQ:
                    switch (GpioPinResistorMode)
                    {
                        // TODO: Pull-up/pull-down modes don't appear to be implemented yet. Disabling this for now.
                        //case PortResistorMode.PullDown:
                        //    pin.SetDriveMode(GpioPinDriveMode.InputPullDown);
                        //    break;

                        //case PortResistorMode.PullUp:
                        //    pin.SetDriveMode(GpioPinDriveMode.InputPullUp);
                        //    break;

                        case PortResistorMode.Disabled:
                        default:
                            pin.SetDriveMode(GpioPinDriveMode.Input);
                            break;
                    }

                    if (GpioPinMode == GoPortGpioPinMode.IRQ)
                    {
                        pin.ValueChanged += GpioPin_ValueChanged;
                    }
                    break;
            }

            _gpioPin = pin;
        }

        private void GpioPin_ValueChanged(GpioPin sender, GpioPinValueChangedEventArgs args)
        {
            // TODO: There is currently an issue where the notified edge is occasionally reported incorrectly.
            // (This seems to happen with pulse signals in particular.)
            // Sending a notification for all edge types for now.
            SendOnGpioPinInterrupt(sender.Read() == GpioPinValue.High);

            //switch (args.Edge)
            //{
            //    case GpioPinEdge.FallingEdge:
            //        switch (GpioPinInterruptMode)
            //        {
            //            case PortInterruptMode.InterruptEdgeLow:
            //            case PortInterruptMode.InterruptEdgeBoth:
            //            case PortInterruptMode.InterruptEdgeLevelLow:
            //                SendOnGpioPinInterrupt(false);
            //                break;
            //        }
            //        break;

            //    case GpioPinEdge.RisingEdge:
            //        switch (GpioPinInterruptMode)
            //        {
            //            case PortInterruptMode.InterruptEdgeHigh:
            //            case PortInterruptMode.InterruptEdgeBoth:
            //            case PortInterruptMode.InterruptEdgeLevelHigh:
            //                SendOnGpioPinInterrupt(true);
            //                break;
            //        }
            //        break;
            //}
        }

        #endregion

        #region SPI

        protected SpiDevice _spi;
        protected SpiDevice _spiEnumeration;
        protected GpioPin _spiCSPin;

        protected virtual SpiDevice GetSpiDevice()
        {
            // If we're bound to a module, return that module's SPI device
            if (BoundModule != null)
                return _spi;

            // Otherwise, return the default enumeration device
            return _spiEnumeration;
        }

        public override void SPIWrite(byte[] writeBuffer)
        {
            if (_spiCSPin != null)
                _spiCSPin.Write(GpioPinValue.Low);

            GetSpiDevice().Write(writeBuffer);

            if (_spiCSPin != null)
                _spiCSPin.Write(GpioPinValue.High);
        }

        public override void SPIWriteRead(byte[] writeBuffer, byte[] readBuffer)
        {
            if (_spiCSPin != null)
                _spiCSPin.Write(GpioPinValue.Low);

            GetSpiDevice().TransferFullDuplex(writeBuffer, readBuffer);

            if (_spiCSPin != null)
                _spiCSPin.Write(GpioPinValue.High);
        }

        public override void SPIWriteRead(byte[] writeBuffer, int writeIndex, int writeLength, byte[] readBuffer, int readIndex, int readLength, int startReadOffset)
        {
            // TODO
            throw new NotImplementedException();
        }

        #endregion

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (!_disposed)
            {
                if (disposing)
                {
                    ReleaseGpioPin();
                    if (_spiEnumeration != null)
                        _spiEnumeration.Dispose();
                    if (_spi != null)
                        _spi.Dispose();
                    if (_spiCSPin != null)
                        _spiCSPin.Dispose();
                }
            }
            _disposed = true;
        }

        ~GoPort()
        {
            Dispose(false);
        }
    }
}
