using System;
using System.Net.Sockets;
using System.Threading;

namespace GoBusExperimental
{
    internal delegate void MessageReceivedEventHandler(MessageType messageType, byte[] messageBytes, int offset, int length);

    internal class CommunicationsChannel
    {
        // Some randomly chosen bytes used to indicate the beginning of a message (to avoid framing errors)
        private static readonly byte[] MessageStartBytes = new byte[] { 0xFF, 0xFE, 0x01, 0xFE, 0xFF };

        private bool _running;
        private readonly Socket _socket;
        private readonly byte[] _sendBuffer = new byte[1000];
        private readonly byte[] _receiveBuffer = new byte[1000];

        public CommunicationsChannel(Socket socket)
        {
            _socket = socket;
        }

        public void Start()
        {
            if (_running)
                return;

            _running = true;
            Thread runThread = new Thread(RunLoop);
            runThread.Start();
        }

        public void Stop()
        {
            _running = false;
        }

        public void SendMessage(MessageType messageType, byte[] messageBytes, int offset, int length)
        {
            Array.Copy(MessageStartBytes, _sendBuffer, MessageStartBytes.Length);
            int index = MessageStartBytes.Length;
            int messageLength = length + 1; // Message Type
            _sendBuffer[index + 0] = (byte)(messageLength >> 8);
            _sendBuffer[index + 1] = (byte)(messageLength >> 0);
            _sendBuffer[index + 2] = (byte)messageType;
            Array.Copy(messageBytes, offset, _sendBuffer, index + 3, length);

            // Calculate new message length
            int totalMessageLength = MessageStartBytes.Length + messageLength + 2;

            try
            {
                _socket.Send(_sendBuffer, 0, totalMessageLength, SocketFlags.None);
            }
            catch { }
        }

        private void RunLoop()
        {
            using (_socket)
            {
                try
                {
                    int bytesRead;
                    int index=0;
                    int currentMessageLength = 0;
                    MessageType currentMessageType = MessageType.None;

                    while(_running)
                    {
                        bytesRead = _socket.Receive(_receiveBuffer, index, 1, SocketFlags.None);
                        if (bytesRead == 0 || !_running)
                            return;

                        // Message Format:
                        //   [Message Start Sequence (Synchronization)]
                        //   [Message Length (from this point on, including message type byte)]
                        //   [Message Type]
                        //   [Message Body]

                        // Message Start
                        if (index < MessageStartBytes.Length)
                        {
                            // Verify that we received the correct byte
                            if (_receiveBuffer[index] != MessageStartBytes[index])
                            {
                                // Invalid, but maybe it's the start of a new sync byte
                                if (_receiveBuffer[index] == MessageStartBytes[0])
                                {
                                    _receiveBuffer[0] = _receiveBuffer[index];
                                    index = 1;
                                }
                                else
                                {
                                    index = 0;
                                }
                            }
                            else
                            {
                                // Correct byte received, carry on...
                                index++;
                            }
                        }
                        // Message length
                        else if (index == MessageStartBytes.Length+0)
                        {
                            currentMessageLength = (_receiveBuffer[index] << 8);
                            index++;
                        }
                        else if (index == MessageStartBytes.Length+1)
                        {
                            currentMessageLength |= (_receiveBuffer[index] << 0);

                            // Is this a zero-byte message?
                            if (currentMessageLength == 0)
                                index = 0;
                            else
                                index++;
                        }
                        // Message Type
                        else if (index == MessageStartBytes.Length +2)
                        {
                            currentMessageType = (MessageType)_receiveBuffer[index];

                            // Is there a message body?
                            if (currentMessageLength == 1)
                                index = 0;
                            else
                                index++;
                        }
                        // Message Body
                        else
                        {
                            // Have we received a full message?
                            if ((index + 1) == MessageStartBytes.Length + 2 + currentMessageLength)
                            {
                                // Notify any listeners
                                if (MessageReceived != null)
                                    MessageReceived(currentMessageType, _receiveBuffer, MessageStartBytes.Length + 2 + 1, currentMessageLength - 1);

                                index = 0;
                            }
                            else if (index + 1 >= _receiveBuffer.Length)
                            {
                                // Buffer overflow
                                index = 0;
                            }
                            else
                            {
                                index++;
                            }
                        }
                    }
                }
                catch { }
            }
        }

        public event MessageReceivedEventHandler MessageReceived;
    }
}
