using System;
using Microsoft.SPOT;
using System.Net;
using System.Net.Sockets;
using System.Threading;

namespace GoBusExperimental
{
    public class NetworkGoHubClient : GoHubBase
    {
        private CommunicationsChannel _channel;
        private readonly byte[] _sendBuffer = new byte[1000];

        private readonly AutoResetEvent _spiWriteReadResponseSignal = new AutoResetEvent(false);
        private readonly byte[] _spiWriteReadResponse = new byte[1000];
        private int _spiWriteReadResponseLength;

        public NetworkGoHubClient(IPAddress serverIP, int port)
        {
            // Connect to the server
            var socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IPEndPoint remoteEndPoint = new IPEndPoint(serverIP, port);
            socket.Connect(remoteEndPoint);

            // Create a new communications channel and start receiving messages
            _channel = new CommunicationsChannel(socket);
            _channel.MessageReceived += CommunicationsChannel_MessageReceived;
            _channel.Start();
        }

        public void AddGoPort(NetworkGoPortClient goPort)
        {
            goPort.Hub = this;
            _goPorts.Add(goPort);
        }

        private void CommunicationsChannel_MessageReceived(MessageType messageType, byte[] messageBytes, int offset, int length)
        {
            try
            {
                switch (messageType)
                {
                    case MessageType.SPIWriteReadResponse:
                        {
                            // We're not currently verifying that the port matches, etc.
                            // Lots of room for improvement, e.g., simultaneous module communications, etc.

                            // Get the port
                            int portID = messageBytes[offset + 0];

                            // Copy the message bytes to our local buffer
                            int byteCount = length - 1;
                            Array.Copy(messageBytes, offset + 1, _spiWriteReadResponse, 0, byteCount);
                            _spiWriteReadResponseLength = byteCount;

                            _spiWriteReadResponseSignal.Set();
                        }
                        break;

                    case MessageType.GpioPinInterrupt:
                        {
                            // Get the port
                            int portID = messageBytes[offset + 0];
                            var port = GetGoPort(portID) as NetworkGoPortClient;
                            if (port == null)
                                return;

                            // Raise the interrupt
                            port.SendOnGpioPinInterrupt(false);
                        }
                        break;
                }
            }
            catch { }
        }

        internal void SPIWrite(NetworkGoPortClient goPort, uint speed, byte[] writeBuffer, int writeIndex, int writeLength)
        {
            // Set up the message
            _sendBuffer[0] = (byte)goPort.ID;
            _sendBuffer[1] = (byte)(speed >> 8);
            _sendBuffer[2] = (byte)(speed >> 0);
            Array.Copy(writeBuffer, writeIndex, _sendBuffer, 3, writeLength);

            // Send the message to the server
            _channel.SendMessage(MessageType.SPIWrite, _sendBuffer, 0, writeLength + 3);
        }

        internal void SPIWriteRead(NetworkGoPortClient goPort, uint speed, byte[] writeBuffer, byte[] readBuffer)
        {
            // Set up the message
            _sendBuffer[0] = (byte)goPort.ID;
            _sendBuffer[1] = (byte)(speed >> 8);
            _sendBuffer[2] = (byte)(speed >> 0);
            Array.Copy(writeBuffer, 0, _sendBuffer, 3, writeBuffer.Length);

            // Send the message to the server
            _channel.SendMessage(MessageType.SPIWriteRead, _sendBuffer, 0, writeBuffer.Length + 3);

            // Wait for a response (to read data back from the module)
            if (_spiWriteReadResponseSignal.WaitOne(500, false))
                Array.Copy(_spiWriteReadResponse, readBuffer, _spiWriteReadResponseLength);
        }

        internal void SetGpioPinMode(NetworkGoPortClient goPort, GoPortGpioPinMode pinMode)
        {
            // Set up the message
            _sendBuffer[0] = (byte)goPort.ID;
            _sendBuffer[1] = (byte)pinMode;

            // Send the message to the server
            _channel.SendMessage(MessageType.GpioPinMode, _sendBuffer, 0, 2);
        }

        public override void OnGoPortBinding(IGoPort goPort)
        {
            base.OnGoPortBinding(goPort);

            // TODO: Tell the hub server that we're binding
        }

        public override void OnGoPortBound(IGoPort goPort)
        {
            base.OnGoPortBound(goPort);

            // TODO: Tell the hub server that we've bound
        }

        public override void OnGoPortUnbound(IGoPort goPort)
        {
            base.OnGoPortUnbound(goPort);

            // TODO: Tell the hub server that we've unbound
        }
    }
}
