﻿using System;
using System.Threading;

namespace GoBusExperimental
{
    public class PiezoBuzzer : GoModule
    {
        Guid _moduleGuid = new Guid(new byte[] { 0x80, 0xA4, 0x99, 0x8D, 0xFF, 0x15, 0x6C, 0x4B, 0x8F, 0xC5, 0xE4, 0xE8, 0x59, 0x75, 0xE3, 0xE6 }); // piezo buzzer
        public override Guid ModuleGuid { get { return _moduleGuid; } }

        AutoResetEvent _irqPortInterruptEvent = new AutoResetEvent(false);

        // 1 byte command, 16 bytes data/GUID, 1 byte CRC8
        private const int _frameLength = 1 + 16 + 1;
        private byte[] _writeFrameBuffer = new byte[_frameLength];
        private byte[] _readFrameBuffer = new byte[_frameLength];

        public PiezoBuzzer()
            :base()
        {
        }

        public PiezoBuzzer(IGoPort goPort)
            :base()
        {
        }

        protected override void Initialize()
        {
            _goPort.GpioPinMode = GoPortGpioPinMode.IRQ;
            _goPort.OnGpioPinInterrupt += _irqPort_OnInterrupt;

            base.Initialize();
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                // clean up any managed code objects
                _goPort.OnGpioPinInterrupt -= _irqPort_OnInterrupt;
            }
            // clean up any unmanaged code objects

            // dispose of our base object
            base.Dispose(disposing);
        }

        void _irqPort_OnInterrupt(object sender, bool state)
        {
            _irqPortInterruptEvent.Set();
        }

        public void SetFrequency(double value)
        {
            int frequency;

            if (value <= 0)
            {
                frequency = 0;
            }
            else
            {
                // The Piezo Buzzer module time base is 1000 kHz
                frequency = (int)(1000000F / value + 0.5);
            }

            int iRetry = 0;
            while (iRetry < 36)
            {
                _writeFrameBuffer[0] = 0x80;
                _writeFrameBuffer[1] = 0x02;  // Module command 'SetColor'
                _writeFrameBuffer[2] = (byte)(frequency >> 8);
                _writeFrameBuffer[3] = (byte)(frequency >> 0);
                _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
                _goPort.SPIWrite(_writeFrameBuffer);

                int verifyFrequency;
                GetFrequency(out verifyFrequency);

                if (frequency == verifyFrequency)
                    return;

                iRetry++;
            }
        }

        private void GetFrequency(out int frequency)
        {
            _writeFrameBuffer[0] = 0x80;
            _writeFrameBuffer[1] = 0x01;  // Module command 'Get frequency'
            for (int i = 2; i < _writeFrameBuffer.Length - 1; i++)
                _writeFrameBuffer[i] = 0x00;
            _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
            _goPort.SPIWrite(_writeFrameBuffer);

            // now, wait for irq to be asserted
            bool responseReceived = false;
            int numRetries = 4;
            int iRetry = 0;
            while (true)
            {
                responseReceived = _irqPortInterruptEvent.WaitOne(10);

                if (responseReceived || (iRetry >= numRetries - 1))
                {
                    // parse and return the response
                    _writeFrameBuffer[0] = 0x00; // get data from module
                    _writeFrameBuffer[1] = 0x00;
                    _writeFrameBuffer[2] = 0x00;
                    _writeFrameBuffer[3] = 0x00;
                    _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
                    _goPort.SPIWriteRead(_writeFrameBuffer, _readFrameBuffer);

                    if (_readFrameBuffer[2] == 2)
                    {
                        frequency = (_readFrameBuffer[3] << 8);
                        frequency |= _readFrameBuffer[4];
                        return;
                    }
                    else if (iRetry >= numRetries - 1)
                    {
                        frequency = -1;
                        return;
                    }
                    else
                    {
                        // failed to receive response; retry
                        responseReceived = false;
                    }
                }
                else
                {
                    // TODO: limit our retries to a handful of retries (or maybe a few dozen -- quickly)
                    // retry infinitely
                    _writeFrameBuffer[0] = 0x80;
                    _writeFrameBuffer[1] = 0x01;  // Module command 'Get frequency'
                    _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
                    _goPort.SPIWrite(_writeFrameBuffer);
                }

                iRetry++;
            }
        }

    }
}
