﻿using System;
using System.Threading;

namespace GoBusExperimental
{
    public class Potentiometer : GoModule
    {
        Guid _moduleGuid = new Guid(new byte[] { 128, 33, 122, 36, 198, 209, 19, 77, 138, 35, 124, 44, 6, 173, 150, 188 }); // potentiometer
        public override Guid ModuleGuid { get { return _moduleGuid; } }

        AutoResetEvent _irqPortInterruptEvent = new AutoResetEvent(false);

        // 1 byte command, 16 bytes data/GUID, 1 byte CRC8
        private const int _frameLength = 1 + 16 + 1;
        private byte[] _writeFrameBuffer = new byte[_frameLength];
        private byte[] _readFrameBuffer = new byte[_frameLength];

        double _lastAdcValue = 0;

        public Potentiometer()
            : base()
        {
        }

        public Potentiometer(IGoPort goPort)
            : base(goPort)
        {
        }

        protected override void Initialize()
        {
            _goPort.GpioPinMode = GoPortGpioPinMode.IRQ;
            _goPort.OnGpioPinInterrupt += _irqPort_OnInterrupt;

            base.Initialize();
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                // clean up any managed code objects
                _goPort.OnGpioPinInterrupt -= _irqPort_OnInterrupt;
            }
            // clean up any unmanaged code objects

            // dispose of our base object
            base.Dispose(disposing);
        }

        void _irqPort_OnInterrupt(object sender, bool state)
        {
            _irqPortInterruptEvent.Set();
        }

        public double GetValue()
        {
            _writeFrameBuffer[0] = 0x80;
            _writeFrameBuffer[1] = 0x01;  // Module command 'Get ADC value'
            _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
            _goPort.SPIWrite(_writeFrameBuffer);

            // now, wait for irq to be asserted
            bool responseReceived = false;
            int numRetries = 36; //4
            int iRetry = 0;
            while (true)
            {
                responseReceived = _irqPortInterruptEvent.WaitOne(10);

                if (responseReceived || (iRetry >= numRetries - 1))
                {
                    // parse and return the response
                    _writeFrameBuffer[0] = 0x00; // get data from module
                    _writeFrameBuffer[1] = 0x00;
                    _writeFrameBuffer[2] = 0x00;
                    _writeFrameBuffer[3] = 0x00;
                    _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
                    _goPort.SPIWriteRead(_writeFrameBuffer, _readFrameBuffer);

                    if (_readFrameBuffer[2] == 2)
                    {
                        // Received two byte ADC value
                        double adcValue = ((short)((_readFrameBuffer[3] << 8) | _readFrameBuffer[4])) / (double)1023;
                        _lastAdcValue = adcValue;
                        return adcValue;
                    }
                    else if (iRetry >= numRetries - 1)
                    {
                        return _lastAdcValue;
                    }
                    else
                    {
                        // failed to receive response; retry
                        responseReceived = false;
                    }
                }
                else
                {
                    // TODO: limit our retries to a handful of retries (or maybe a few dozen -- quickly)
                    // retry infinitely
                    _writeFrameBuffer[0] = 0x80;
                    _writeFrameBuffer[1] = 0x01;  // Module command 'Get ADC value'
                    _writeFrameBuffer[_writeFrameBuffer.Length - 1] = CRC8.Compute8(_writeFrameBuffer, 0, _writeFrameBuffer.Length - 1);
                    _goPort.SPIWrite(_writeFrameBuffer);
                }

                iRetry++;
            }
        }
    }
}
