using System;

namespace GoBusExperimental
{
    public delegate void GpioEventHandler(object sender, bool state);

    public static class GoBusUtility
    {
#if NETMF
        public static Microsoft.SPOT.Hardware.Port.ResistorMode ToNETMF(this PortResistorMode mode)
        {
            switch (mode)
            {
                case PortResistorMode.Disabled:
                    return Microsoft.SPOT.Hardware.Port.ResistorMode.Disabled;
                case PortResistorMode.PullDown:
                    return Microsoft.SPOT.Hardware.Port.ResistorMode.PullDown;
                case PortResistorMode.PullUp:
                    return Microsoft.SPOT.Hardware.Port.ResistorMode.PullUp;
            }

            throw new ArgumentException("mode");
        }

        public static Microsoft.SPOT.Hardware.Port.InterruptMode ToNETMF(this PortInterruptMode mode)
        {
            switch (mode)
            {
                case PortInterruptMode.InterruptNone:
                    return Microsoft.SPOT.Hardware.Port.InterruptMode.InterruptNone;
                case PortInterruptMode.InterruptEdgeLow:
                    return Microsoft.SPOT.Hardware.Port.InterruptMode.InterruptEdgeLow;
                case PortInterruptMode.InterruptEdgeHigh:
                    return Microsoft.SPOT.Hardware.Port.InterruptMode.InterruptEdgeHigh;
                case PortInterruptMode.InterruptEdgeBoth:
                    return Microsoft.SPOT.Hardware.Port.InterruptMode.InterruptEdgeBoth;
                case PortInterruptMode.InterruptEdgeLevelHigh:
                    return Microsoft.SPOT.Hardware.Port.InterruptMode.InterruptEdgeLevelHigh;
                case PortInterruptMode.InterruptEdgeLevelLow:
                    return Microsoft.SPOT.Hardware.Port.InterruptMode.InterruptEdgeLevelLow;
            }

            throw new ArgumentException("mode");
        }

        public static bool WaitOne(this System.Threading.WaitHandle handle, int millisecondsTimeout)
        {
            return handle.WaitOne(millisecondsTimeout, false);
        }
#endif

        // Adapted from the official GoBus library. TODO: Cleanup/rework
        public static Guid GetSPIConnectedModuleGuid(IGoPort goPort)
        {
            int frameLength = 1 + 16 + 1;
            byte[] writeFrameBuffer = new byte[frameLength];
            byte[] readFrameBuffer = new byte[frameLength];
            //SPI.Configuration spiConfig = new SPI.Configuration(spiChipSelectPin, false, 0, 0, false, false, 500, spiModule);
            int iEnumAttempt = 0;
            int emptyFrameCount = 0;
            byte basicDeviceType = 0x00;
            //using (SPI spi = new SPI(spiConfig))
            {
                while (iEnumAttempt < 72) // NOTE: 72 attempts allow our frame to rotate four full times in case of frame mis-alignment; this should be unnecessary in time.
                {
                    writeFrameBuffer[0] = 0xFE; // 'Enumerate' command
                    for (int i = 1; i < writeFrameBuffer.Length - 1; i++)
                        writeFrameBuffer[i] = 0xFF; // filler -- so that module sends ID
                    writeFrameBuffer[writeFrameBuffer.Length - 1] = CRC8.Compute8(writeFrameBuffer, 0, writeFrameBuffer.Length - 1);
                    //spi.WriteRead(writeFrameBuffer, readFrameBuffer);
                    //goPort.SPIWriteRead(spiConfig, writeFrameBuffer, readFrameBuffer);
                    goPort.SPIWriteRead(writeFrameBuffer, readFrameBuffer);

                    // validate the response...
                    bool replyIsBlank = true;
                    for (int i = 1; i < readFrameBuffer.Length - 1; i++)
                    {
                        if (readFrameBuffer[i] != 0)
                            replyIsBlank = false;
                    }

                    if (replyIsBlank)
                    {
                        // no response; try again (up to 4 times total)
                        emptyFrameCount++;
                        if (emptyFrameCount > 4)
                            return Guid.Empty;
                    }
                    else if (CRC8.Compute8(readFrameBuffer) == 0)
                    {
#if NETMF
                        return new Guid(Microsoft.SPOT.Hardware.Utility.ExtractRangeFromArray(readFrameBuffer, 1, readFrameBuffer.Length - 2));
#else
                        byte[] guidBytes = new byte[16];
                        Buffer.BlockCopy(readFrameBuffer, 1, guidBytes, 0, 16);
                        return new Guid(guidBytes);
#endif
                    }
                    else if (readFrameBuffer[1] != 0)
                    {
                        bool replyIsBasicDevice = true;
                        // did not pass CRC; see if it's a basic device type (first byte is id, then all other bytes including CRC will be zeros)
                        // Note: on WinIoT SPI works a bit differently -- the MOSI line goes low between each byte, resetting the shift register on basic modules.
                        // Because of this we also need to allow for each byte being the ID
                        for (int i = 2; i < readFrameBuffer.Length; i++)
                        {
                            if (readFrameBuffer[i] != 0 && readFrameBuffer[i] != readFrameBuffer[1])
                            {
                                replyIsBasicDevice = false;
                                break;
                            }
                        }

                        if (replyIsBasicDevice)
                        {
                            if (basicDeviceType == 0)
                            {
                                // first successful query; store this id and we'll verify by second enum
                                basicDeviceType = readFrameBuffer[1];
                            }
                            else
                            {
                                if (basicDeviceType == readFrameBuffer[1])
                                {
                                    // Make sure the other bytes are zero
                                    for (int i = 2; i < readFrameBuffer.Length; i++)
                                        readFrameBuffer[i] = 0;

#if NETMF
                                    return new Guid(Microsoft.SPOT.Hardware.Utility.ExtractRangeFromArray(readFrameBuffer, 1, readFrameBuffer.Length - 2));
#else
                                    byte[] guidBytes = new byte[16];
                                    Buffer.BlockCopy(readFrameBuffer, 1, guidBytes, 0, 16);
                                    return new Guid(guidBytes);
#endif
                                }
                                else
                                {
                                    // mismatch; could not enumerate successfully
                                    return Guid.Empty;
                                }
                            }
                        }
                    }
                    else
                    {
                        // corrupt message; try again
                    }

                    iEnumAttempt++;
                }
            }

            return Guid.Empty;
        }
    }
}
