using System;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

namespace GoBusExperimental
{
    public class GoPort : GoPortBase
    {
        public GoPort(int id, Cpu.Pin gpioPin, Cpu.Pin spiChipSelectPin, SPI.SPI_module spiModule, Cpu.Pin uartTXPin, Cpu.Pin uartRXPin)
        {
            ID = id;
            _gpioPin = gpioPin;
            _spiChipSelectPin = spiChipSelectPin;
            _spiModule = spiModule;
            _uartTXPin = uartTXPin;
            _uartRXPin = uartRXPin;

            _spiEnumerationConfig = new SPI.Configuration(spiChipSelectPin, false, 0, 0, false, false, GoBus.SPIEnumerationClockRate, spiModule);
        }

        public override void Initialize()
        {
            // Make sure the output pin is high so the module doesn't inadvertently respond to another port's enumeration request
            OutputPort csPort = new OutputPort(_spiChipSelectPin, true);
            csPort.Dispose();

            // Set TX and RX pins to input, pull-up mode
            InputPort inputPort;
            if (_uartTXPin >= 0)
            {
                inputPort = new InputPort(_uartTXPin, false, Port.ResistorMode.PullUp);
                inputPort.Dispose();
            }
            if (_uartRXPin >= 0)
            {
                inputPort = new InputPort(_uartRXPin, false, Port.ResistorMode.PullUp);
                inputPort.Dispose();
            }
        }

        protected override void OnBinding(GoModule module)
        {
            // Initialize the GPIO pin
            InitializeGpioPin();

            // Set up SPI configuration
            SPI.Configuration spiConfig = new SPI.Configuration(_spiChipSelectPin, false, 0, 0, false, false, module.SPIClockRate, _spiModule);
            _spiConfig = spiConfig;

            // Does this module need UART?
            var uartModule = module as IUartGoModule;
            if (uartModule != null)
            {
                // TODO
                throw new NotImplementedException();
            }
        }

        protected override void OnUnbinding()
        {
            // Release the GPIO pin
            ReleaseGpioPin();

            if (_spiConfig != null)
            {
                _spiConfig = null;
            }
        }

        #region SPI

        protected readonly Cpu.Pin _spiChipSelectPin;
        protected readonly SPI.SPI_module _spiModule;
        protected SPI.Configuration _spiConfig;
        protected readonly SPI.Configuration _spiEnumerationConfig;

        protected virtual SPI.Configuration GetSPIConfiguration()
        {
            // If we're bound to a module, return that module's SPI configuration
            if (BoundModule != null)
                return _spiConfig;

            // Otherwise, return the default enumeration configuration
            return _spiEnumerationConfig;
        }

        public override void SPIWriteRead(byte[] writeBuffer, int writeIndex, int writeLength, byte[] readBuffer, int readIndex, int readLength, int startReadOffset)
        {
            using (var spi = new SPI(GetSPIConfiguration()))
                spi.WriteRead(writeBuffer, writeIndex, writeLength, readBuffer, readIndex, readLength, startReadOffset);
        }

        #endregion

        #region UART

        private readonly Cpu.Pin _uartTXPin;
        private readonly Cpu.Pin _uartRXPin;

        #endregion

        #region GPIO

        private readonly Cpu.Pin _gpioPin;
        private bool _gpioOutputValue;
        private OutputPort _gpioOutputPort;
        private InputPort _gpioInputPort;
        private InterruptPort _gpioInterruptPort;

        private GoPortGpioPinMode _gpioPinMode = GoPortGpioPinMode.None;
        public override GoPortGpioPinMode GpioPinMode
        {
            get { return _gpioPinMode; }
            set
            {
                if (_gpioPinMode == value)
                    return;
                _gpioPinMode = value;
                InitializeGpioPin();
            }
        }

        private bool _gpioPinGlitchFilter;
        public override bool GpioPinGlitchFilter
        {
            get { return _gpioPinGlitchFilter; }
            set
            {
                if (_gpioPinGlitchFilter == value)
                    return;
                _gpioPinGlitchFilter = value;
                InitializeGpioPin();
            }
        }

        private PortResistorMode _gpioPinResistorMode = PortResistorMode.PullUp;
        public override PortResistorMode GpioPinResistorMode
        {
            get { return _gpioPinResistorMode; }
            set
            {
                if (_gpioPinResistorMode == value)
                    return;
                _gpioPinResistorMode = value;
                InitializeGpioPin();
            }
        }

        private PortInterruptMode _gpioPinInterruptMode = PortInterruptMode.InterruptEdgeLow;
        public override PortInterruptMode GpioPinInterruptMode
        {
            get { return _gpioPinInterruptMode; }
            set
            {
                if (_gpioPinInterruptMode == value)
                    return;
                _gpioPinInterruptMode = value;
                InitializeGpioPin();
            }
        }

        protected virtual void ReleaseGpioPin()
        {
            // Dispose any existing instances of the GPIO pin
            if (_gpioOutputPort != null)
            {
                _gpioOutputPort.Dispose();
                _gpioOutputPort = null;
            }

            if (_gpioInputPort != null)
            {
                _gpioInputPort.Dispose();
                _gpioInputPort = null;
            }

            if (_gpioInterruptPort != null)
            {
                _gpioInterruptPort.OnInterrupt -= GPIOInterruptPort_OnInterrupt;
                _gpioInterruptPort.Dispose();
                _gpioInterruptPort = null;
            }
        }

        protected virtual void InitializeGpioPin()
        {
            ReleaseGpioPin();

            switch (_gpioPinMode)
            {
                case GoPortGpioPinMode.None:
                    // Nothing to do, we already released resources
                    break;

                case GoPortGpioPinMode.Output:
                    _gpioOutputPort = new OutputPort(_gpioPin, _gpioOutputValue);
                    break;

                case GoPortGpioPinMode.Input:
                    _gpioInputPort = new InputPort(_gpioPin, _gpioPinGlitchFilter, _gpioPinResistorMode.ToNETMF());
                    break;

                case GoPortGpioPinMode.IRQ:
                    _gpioInterruptPort = new InterruptPort(_gpioPin, _gpioPinGlitchFilter, _gpioPinResistorMode.ToNETMF(), _gpioPinInterruptMode.ToNETMF());
                    _gpioInterruptPort.OnInterrupt += GPIOInterruptPort_OnInterrupt;
                    break;
            }
        }

        public override void GpioWrite(bool value)
        {
            if (_gpioOutputPort == null)
                throw new Exception();

            _gpioOutputValue = value;
            _gpioOutputPort.Write(value);
        }

        public override bool GpioRead()
        {
            if (_gpioInputPort != null)
                return _gpioInputPort.Read();

            if (_gpioInterruptPort != null)
                return _gpioInterruptPort.Read();

            throw new Exception();
        }

        private void GPIOInterruptPort_OnInterrupt(uint pin, uint state, DateTime time)
        {
            bool s = !(state == 0);
            SendOnGpioPinInterrupt(s);
        }

        #endregion
    }
}
